/*
 * File:   recollViewer.cpp
 * Author: david
 *
 * Created on November 16, 2011, 10:36 AM
 */

#include <plasma/abstractrunner.h>

#include <QDate>
#include <QDebug>

#include <kicon.h>
#include <krun.h>
#include <Qt/qstringlist.h>

#include "recollViewer.h"
#include "cmdstrings.h"

// The following tags for annotating keywords:
static const QString TAGS[] = { "<font color=blue>", "<font color=green>",
                "<font color=red>", "<font color=brown>", "<font color=violet>" };
static const QString UNTAGS[] = { "</font>", "</font>", "</font>", "</font>", "</font>" };
static const int NTAGS = 5;

recollViewer::annotator::annotator() {
    // assemble pretags that contain no like text pattern
    for ( int n = 0; n < NTAGS; ++n )  {
        QString qn( n );
        m_pretags << "<!" + qn + "!>";
        m_preuntags << "</!" + qn + "!>";
    }
}

QString recollViewer::annotator::annotate( QString orig, QString keyword ) {
    QStringList keywords;
    keywords << keyword;
    return annotate( orig, keywords );
}

QString recollViewer::annotator::annotate( QString orig, QStringList keywords ) {
    // Instead of using/writing a heavy HTML/XTML parser-surgeon, do this in two stages
    // so text inside the tags is protected. Simple, down and dirty!
    QString result = orig;

    int nkw = keywords.count();
    int ntags = m_pretags.count();

    for ( int kw = 0; kw < nkw; ++kw ) {
        //result.replace( keywords[ kw ], m_pretags[ kw % ntags ] + keywords[ kw ]
        //                        + m_preuntags[ kw % ntags ], Qt::CaseInsensitive );
        // prefer options to keep case of original in case recoll preserves it
        int pos = -1;
        while ( ( pos = result.lastIndexOf( keywords[ kw ], pos, Qt::CaseInsensitive )) >= 0 ) {
            // do this backwords:
            result.insert( pos + keywords[ kw ].length(), m_preuntags[ kw % ntags ] );
            result.insert( pos, m_pretags[ kw % ntags ] );
        }
    }

    for ( int tag = 0; tag < ntags; ++tag ) {
        result.replace( m_preuntags[ tag ], UNTAGS[ tag ] );
        result.replace( m_pretags[ tag ], TAGS[ tag ] );
    }

    return result;
}

recollViewparser::recollViewparser( Plasma::QueryMatch* match ) {
    QStringList data = match->data().toStringList();

    m_viewPath = data[ 0 ];
    data.pop_front();
    copyData( data );

    if (m_viewPath.indexOf(EXECCMD, 0) >= 0) {
        m_viewTitle = match->subtext(); // run query so just use
        return;
    }

    if ( hasTitle() ) {
        m_viewTitle = title();
        return;
    }

    if ( hasIPath() ) {
        if ( ipathIsNumber() ) {
            m_viewTitle = recollFileName();
            m_viewTitle += ':';
            m_viewTitle += ipathToNumber();
        }
        else if ( ipathIsNumberColonNumber() ) {
            m_viewTitle = recollFileName();
            m_viewTitle += '/';
            m_viewTitle += ipath(); // shown this way in krunner
        }
        else {
            m_viewTitle = ipath(); // just show item title
        }
        return;
    }

    m_viewTitle = recollFileName();
}
recollViewparser::recollViewparser() {
    m_viewPath = "";
    m_viewTitle =  i18n( "NO MATCHES!");
    
    QStringList list;
    copyData( list );
}

recollViewer::recollViewer() {
    widget.setupUi(this);

    widget.toolButtonPrev->setIcon( KIcon( "go-previous-view-page"));
    widget.toolButtonNext->setIcon( KIcon( "go-next-view-page"));
    widget.toolButtonOpen->setIcon( KIcon( "document-open"));
    widget.toolButtonDone->setIcon( KIcon( "dialog-close" ));
    widget.toolButtonList->setIcon( KIcon( "view-list-details" ) );
    widget.toolButtonResearch->setIcon( KIcon( "run-build-configure" ));

    connect( widget.toolButtonNext, SIGNAL( clicked() ), this, SLOT( tonext() ) );
    connect( widget.toolButtonPrev, SIGNAL( clicked() ), this, SLOT( toprev() ) );
    connect( widget.toolButtonOpen, SIGNAL( clicked() ), this, SLOT( runIt() ) );
    connect( widget.toolButtonResearch, SIGNAL( clicked() ), this, SLOT( requestResearch() ));

    // make floating widget:
    m_klistSelections = new myQListWidget();
    m_klistSelections->setVisible( false );
    widget.toolButtonList->setChecked( false );
    m_bListChecked = false;
    m_klistSelections->setWindowTitle( "Recoll Selections" );
    m_klistSelections->setGeometry(QRect(80, 360, 293, 299));
    m_klistSelections->setFrameShape(QFrame::Panel);
    //m_klistSelections->setLineWidth(1);
    m_klistSelections->setEditTriggers(QAbstractItemView::CurrentChanged | QAbstractItemView::SelectedClicked);
    m_klistSelections->setProperty("showDropIndicator", QVariant(false));
    //m_klistSelections->setAlternatingRowColors(true);
    m_klistSelections->setMovement(QListView::Free);
    m_klistSelections->setSelectionRectVisible(true);
    m_klistSelections->setAutoScroll( true );
    m_klistSelections->setWordWrap( true );
    //m_klistSelections->setUniformItemSizes( true ); 

    connect(widget.toolButtonList, SIGNAL(toggled(bool)), m_klistSelections, SLOT(setVisible(bool)));
    connect(widget.toolButtonDone, SIGNAL(clicked()), m_klistSelections, SLOT(hide()));
    connect(widget.toolButtonDone, SIGNAL(clicked()), this, SLOT(clear()));
    connect(m_klistSelections, SIGNAL(closing()), this, SLOT(closingList()));
    connect(m_klistSelections, SIGNAL(itemSelectionChanged()),this, SLOT(toItem()));

    m_nMatches = 0;
    m_nIndex = 0;
    m_bReopened = false;
}

void recollViewer::clear() {
    m_klistSelections->clear();
    m_klistSelections->reset();
    m_ids.clear();
    m_matches.clear();
}

void recollViewer::runView( const Plasma::RunnerContext& context,
                                const Plasma::QueryMatch& match, RRSettings* options ) {
    clear();

    m_thisMatch = new Plasma::QueryMatch( match );
    
    widget.frame->setWindowIcon( match.runner()->icon() );
    m_klistSelections->setWindowIcon( match.runner()->icon() );  
    
    m_query = context.query();
    m_options = options;
    m_nMatches = context.matches().count();
    if  ( m_nMatches > 0 ) {
        //assemble ids since other find schemes will not work!
        for (int n = 0; n < m_nMatches; ++n) {
            if (context.matches().at(n).runner() == match.runner()) {
                stuffList(context.matches().at(n));
            }
        }
        m_nMatches = m_matches.count(); // how many are OK?
    }
    
    if  ( m_nMatches > 0 ) {
        if  ( match.data().isNull() ) {
            m_nIndex = 0;
        }
        else {
            m_nIndex = m_ids.indexOf( match.id(), 0);
            if (m_nIndex < 0)
                m_nIndex = 0;
        }
        
        m_klistSelections->setCurrentRow(m_nIndex);

        qDebug() << "Starting viewer " << m_nIndex << "/" << m_nMatches;

        stuffView();
    }
    else {
        m_nIndex = 0;
        
        qDebug() << "Starting view no matches";
        stuffNullView();
    }
    
    show();
}

void recollViewer::show() {
    widget.toolButtonList->setChecked( m_bListChecked );
    m_klistSelections->setVisible( m_bListChecked );

    if  ( m_bReopened ) {
        setGeometry( m_saveGeometry );
    }
    m_bReopened = true;
    
    QDialog::show();
}

void recollViewer::stuffList( Plasma::QueryMatch match ) {
    m_ids << match.id();
    m_matches << match;

    recollViewparser parser( &match );

    QListWidgetItem* item = new QListWidgetItem();
    item->setIcon( match.icon() );
    item->setText( parser.viewTitle() );
    m_klistSelections->addItem( item );
}

void recollViewer::stuffView() {
    if (!m_thisMatch->isValid())
        return;
    
    recollViewparser parser(m_thisMatch);

    m_path = parser.viewPath();

    if (m_path.indexOf(EXECCMD, 0) >= 0) {
        widget.linePath->setText("");
        widget.lineMimetype->setText("");
        widget.lineRelevancy->setText("");
        //widget.lineDate->setText("");
    } else {
        widget.linePath->setText(m_thisMatch->subtext());
        widget.lineMimetype->setText(parser.mimeType());
        widget.lineRelevancy->setText(parser.recollRelevStr());
        //widget.lineDate->setText(parser.dateLocaleLongString());
    } 
    
    
    stuffView( parser );
    
    widget.mimeIcon->setIcon( m_thisMatch->icon() );
    widget.toolButtonNext->setDisabled(m_nIndex >= m_nMatches - 1);
    widget.toolButtonPrev->setDisabled(m_nIndex <= 0);
    widget.toolButtonOpen->setEnabled( m_thisMatch->isEnabled());
}

void recollViewer::stuffNullView() {
    m_path = "";
    
    widget.linePath->setText("");
    widget.lineMimetype->setText("");
    widget.lineRelevancy->setText("");
    //widget.lineDate->setText("");

    recollViewparser parser;
    stuffView(parser);
    
    widget.mimeIcon->setIcon( KIcon( "dialog-warning") );
    widget.toolButtonNext->setDisabled( true );
    widget.toolButtonPrev->setDisabled( true );
    widget.toolButtonOpen->setDisabled( true );
}

void recollViewer::stuffView(recollViewparser& parser) {
    const QString TITLESTYLE = "<center>";
    const QString TITLEUNSTYLE = "</center>";

    widget.textTitle->setText(TITLESTYLE
            + annot.annotate(parser.viewTitle(), m_options->m_bFilenamesOnly
                                                 ? m_keywords
                                                 : m_titlekeywords)
            + TITLEUNSTYLE);

    widget.lineAuthor->setText(parser.author());
    widget.textAbstract->setText(annot.annotate(parser.abstract(), m_keywords));
    widget.lineDate->setText(parser.dateLocaleLongString());
}

void recollViewer::tonext() {
    if  ( m_nIndex >= m_nMatches - 1 )
        return;
    
    ++m_nIndex;
    stuffItem();
}

void recollViewer::toprev() {
    if  ( m_nIndex <= 0 )
        return;
    
    --m_nIndex;
    stuffItem();
}

void recollViewer::toItem() {
    toItem( m_klistSelections->currentRow() );
}

void recollViewer::stuffItem() {
    /*QListWidgetItem* item = m_klistSelections->item( m_nIndex );
    m_klistSelections->scrollToItem( item );
    m_klistSelections->setItemSelected( item, true );*/

    m_klistSelections->setCurrentRow( m_nIndex );

    delete m_thisMatch;
    m_thisMatch = new Plasma::QueryMatch(m_matches.at( m_nIndex ));
    stuffView();    
}

void recollViewer::toItem( int item ) {
    m_nIndex = item;
    stuffItem();
}

void recollViewer::runIt() {
    qDebug() << m_path;

    if  ( m_path.indexOf( EXECCMD ) == 0) {
        // should have done this for me!
        KRun::runCommand( m_path.mid( EXECCMD.length() ), 0 );
    }
    else {
        (void) new KRun( m_path, 0 );
    }
}

void recollViewer::closeEvent(QCloseEvent *event)
 {
    qDebug() << "closing viewer";

    m_saveGeometry = geometry();

    m_klistSelections->close();
    event->accept();
 }

void recollViewer::closingList() {
    m_bListChecked = widget.toolButtonList->isChecked();
    widget.toolButtonList->setChecked( false );
}

void recollViewer::requestResearch() {
    emit research( m_query );
}

bool recollViewer::isQueryField(QString code, int item) {
    if (!m_keywords[ item ].startsWith(code))
        return false;

    m_titlekeywords << m_keywords[ item ].mid(code.length() );
    m_keywords.removeAt(item);
    return true;

}

bool recollViewer::isSkipField(QRegExp code, int item) {
    if (!m_keywords[ item ].contains(code))
        return false;

    m_keywords.removeAt(item);
    return true;

}

void recollViewer::registerUserQuery(QString query) {
    m_keywords.clear();         // not necessary
    m_titlekeywords.clear();    // necessary!

    // create list out of spaced entries, leave empties!
    m_keywords = query.split( ' ', QString::KeepEmptyParts );
    // do not care about precedence, recoll does it.
    m_keywords.replaceInStrings( RECOLLPRECEDENCE, "" );

    int kq = -1;        // in quoted sequence, -1 NOT
    int nkw = 0;        // current keyword

    while ( nkw < m_keywords.count() ) {
        // service quotes first
        if  ( kq >= 0 ) {
            // concatenate with space
            m_keywords[ kq ] += ' ';
            m_keywords[ kq ] += m_keywords[ nkw ];
            // ends it?
            if  ( m_keywords[ nkw ].endsWith( RECOLLQUOTE )) {
                m_keywords[ kq ].remove( RECOLLQUOTE );
                kq = -1;
            }
            // done with this entry, period
            m_keywords.removeAt( nkw );
            continue;
        }
        // start new sequence?
        if  ( m_keywords[ nkw ].startsWith( RECOLLQUOTE ) ) {
            // improper but acceptable single word?
            if  ( !m_keywords[ nkw].endsWith( RECOLLQUOTE ))
                kq = nkw;
            m_keywords[ nkw ].remove( RECOLLQUOTE );
            // next
            ++nkw;
            continue;
        }

        // this one is separate, allow these cases since displays overlap logically
        if  ( isQueryField( RECOLLUSERTITLEPREFIX, nkw ) )
            continue;
        if  ( isQueryField( RECOLLUSERFILENAMEPREFIX, nkw ) )
            continue;
        if  ( isQueryField( RECOLLUSERIPATHPREFIX, nkw ) )
            continue;

        // these are simply skipped
        if  ( isSkipField( BLANK, nkw ) )
            continue;
        if  ( isSkipField( RECOLLUSERNOTFLAG, nkw ) )
            continue;
        if  ( isSkipField( RECOLLUSERFIELDFLAG, nkw ) )
            continue;
        if  ( isSkipField( RECOLLANDCMD, nkw ) )
            continue;
        if  ( isSkipField( RECOLLORCMD, nkw ) )
            continue;
        if  ( isSkipField( RECOLLANDNOTCMD, nkw ) )
            continue;

        ++nkw;
    }
}

#include "recollViewer.moc"