/* Async Worker Source Code


 */

#include <QDebug>

#include <kmimetype.h>
#include <kfileitem.h>
#include "recollworker.h"
#include "cmdstrings.h"
#include "recollparser.h"

RecollAsyncWorker::RecollAsyncWorker(  Plasma::AbstractRunner* runner,
        Plasma::RunnerContext& context,
        QIcon icon ) {
    setProcessChannelMode(QProcess::SeparateChannels);
    connect( this, SIGNAL( readyReadStandardOutput()),
            this, SLOT( processRecollOutput()));

    m_runner = runner;
    m_context = context;
    m_term = context.query();
    m_icon = icon;
}

void RecollAsyncWorker::setOptions( RRSettings* options ) {
    m_options = options;

    m_fileTypes = m_options->m_fileTypes.split( QRegExp( "[\\s,:]+" ) );
    m_filePatterns = m_options->m_filePatterns.split( QRegExp( "[\\s,:]+" ) );
}

void RecollAsyncWorker::startWorker( QString& cmd, QStringList& args,
                QStringList& abbrev) {
    m_nHits = 0;
    
    m_recollArgs = abbrev;

    postProcess();
    
    start( cmd, args );
    //echoCmd( cmd, args );

}

void RecollAsyncWorker::processRecollOutput() {
    QByteArray bytes = readAllStandardOutput();       // flush even if
    //qDebug() << "readAllStandardOutput():" << bytes;
    
    if (m_options->m_bRestrictHits && m_nHits >= m_options->m_nMaxHits) {
        // this can occur so ....
        //kill();
        return;
    }
    QString qsMimetype;
    QString qsMessage;    

    QList <QByteArray> Lines = bytes.split('\n');
    //qDebug() << "raw count:" << Lines.count();
    
    if  ( Lines.count() < 1 ) //no lines separated, so use ouptut as whole
        Lines << bytes;

    //qDebug() << "adj count" << Lines.count();

    int UseMax = m_options->m_bRestrictHits
                        ? m_options->m_nMaxHits
                        : 0;
    
    for (int ln = 0; ln < Lines.count(); ++ln) {
        //qDebug() << Lines[ ln ];
        recollparser* parser = new recollparser( Lines[ ln ], UseMax );
        if  ( !parser->isValid() )
            continue;

        if  ( m_options->m_bRestrictRank
                && parser->recollRelevInt() < m_options->m_nMinRank )
            continue;
        
        // now extract 
        qsMimetype = parser->mimeType();
        
        // DEBUG
        //echoStr( Lines[ ln ] );
        // Second [...] will have a file name, will differ if email!
        qsMessage = "";
        bool legit = parser->isFile();
        if ( m_options->m_bFilenamesOnly )
            legit = legit || parser->isDir();
        if ( !m_options->m_bNonFiles && !legit)
            continue;
        
        if (!m_options->m_bHiddenFiles) {
            if (parser->isHidden())
                continue;
            // what if the directory is hidden?? Apparently
            // qfile.isHidden() does not pick this up!
            if (parser->absolutePath().indexOf("/.") >= 0)
                continue;
        }
         
        QString path = parser->absolutePath();

        if  ( m_options->m_bNoArchives ) {
            KFileItem info( KFileItem::Unknown, KFileItem::Unknown, path );
            if  ( info.mimeComment().contains( "archive", Qt::CaseInsensitive ) )
                continue;
        }
        
        if  ( m_options->m_bRestrictTypes && !matchAnyType( qsMimetype ) )
            continue;
        if  ( m_options->m_bRestrictPatterns && !( m_options->m_bAllPatterns
                                            ? matchAllPattern( parser->file() )
                                            : matchAnyPattern( parser->file() ) )
                )
            continue;

        Plasma::QueryMatch myMatch(m_runner);
        myMatch.setType(Plasma::QueryMatch::HelperMatch);
        
        QStringList cmdData;
        cmdData << parser->urlPath();
        cmdData << parser->data();
        myMatch.setId( QString( m_nHits ) );
        myMatch.setData( cmdData );


        QString qsText;
        if  ( m_options->m_bTag )
            qsText += RECOLLTAG;
        qsText += parser->title();

        if  ( !parser->hasTitle() ) {
            // title itself is what belongs.
            // if this is the filename, then, some tricks
            if  ( parser->hasIPath() ) {
                if  ( parser->ipathIsNumber() ) {
                    qsText += ':';
                    qsText += parser->ipathToNumber();
                }
                else {
                    qsText += '/';
                    qsText += parser->ipath();
                }
            }
        }

        if  ( m_options->m_bTag ) {
            qsText += " (";
            qsText += parser->recollRelevStr();
            qsText += ")";
        }
        
        myMatch.setText( qsText );
        myMatch.setSubtext( path );
        myMatch.setEnabled( legit );
 
        // try using parsed mimetype
        if ( iconCache.contains( parser->mimeType() )) {
            myMatch.setIcon(iconCache.value( parser->mimeType() ));
        }
        else if ( iconCache.contains( parser->urlPath() )) {
            myMatch.setIcon(iconCache.value( parser->urlPath() ));
        }
        else {
            KIcon icon;
            /*KIcon icon( legit
                        ? KMimeType::iconNameForUrl(parser->urlPath())
                        : qsMimetype.replace( '/', '-' )
            );*/
            if  ( parser->mimeType() != "" ) {
                icon = KIcon( parser->mimeType().replace( QRegExp( "[\\/\\.]" ), "-" ) );
                iconCache.insert( parser->mimeType(), icon);
            }
            else {
                // get it from the file
                icon = KIcon( KMimeType::iconNameForUrl( parser->urlPath() ) );
                iconCache.insert( parser->urlPath(), icon);
            }
            myMatch.setIcon(icon);
        }
        
        float relev;
        if  ( m_options->m_sortType != RANK ) {
            if (!legit) {
                relev = parser->mapSeqToRange( m_nHits, RELEV_NONFILE, RELEV_FILE);
            } else if (parser->isDir()) {
                relev = parser->mapSeqToRange(m_nHits, RELEV_DIRECTORY, RELEV_EXEC);
            } else {
                relev = parser->mapSeqToRange(m_nHits, RELEV_FILE, RELEV_DIRECTORY);
            }

        } else {
            if (!legit) {
                relev = parser->mapRelevToRange(RELEV_NONFILE, RELEV_FILE);
            } else if (parser->isDir()) {
                relev = parser->mapRelevToRange(RELEV_DIRECTORY, RELEV_EXEC);
            } else {
                relev = parser->mapRelevToRange(RELEV_FILE, RELEV_DIRECTORY);
            }
        }
        myMatch.setRelevance( relev );
        
        m_context.addMatch(m_term, myMatch);

        ++m_nHits; 
        if  ( m_options->m_bRestrictHits && m_nHits >= m_options->m_nMaxHits ) {
            kill();
            break;
        }
    }
}

bool RecollAsyncWorker::matchAnyType( QString& mime ) {
    for ( int n = 0; n < m_fileTypes.count(); ++n ) {
        if  ( matchOneType( mime, m_fileTypes[ n ] ) )
            return true;
    }
    return false;   // none matched
}

bool RecollAsyncWorker::matchOneType( QString& mime, QString& type ) {
    // case independent
    // test group/subgroup
    if  ( type.compare( mime, Qt::CaseInsensitive ) == 0 )
        return true;

    // test individual groups, subgroups
    QStringList groupsubgroup = mime.split( '/' );
    for (int n = 0; n < groupsubgroup.count(); ++n ) {
        if  ( type.compare( groupsubgroup[ n ], Qt::CaseInsensitive ) == 0 )
            return true;
    }
    return false;       // no match
}

bool RecollAsyncWorker::matchAllPattern( const QFileInfo& finfo ) {
    for ( int n = 0; n < m_filePatterns.count(); ++n ) {
        if  ( !matchOnePattern( finfo, m_filePatterns[ n ] ) )
            return false;
    }
    return true;   // none matched
}
bool RecollAsyncWorker::matchAnyPattern( const QFileInfo& finfo ) {
    for ( int n = 0; n < m_filePatterns.count(); ++n ) {
        if  ( matchOnePattern( finfo, m_filePatterns[ n ] ) )
            return true;
    }
    return false;   // none matched
}

bool RecollAsyncWorker::matchOnePattern( const QFileInfo& finfo , QString& pattern ) {
#ifdef Q_OS_WIN
    char sep = '\\';
    Qt::CaseSensitivity cs = Qt::CaseInsensitive;
#else
    char sep = '/';
    Qt::CaseSensitivity cs = Qt::CaseSensitive;
#endif

    // test extension .ext
    int pos = pattern.indexOf( '.' );
    if  ( pos >= 0 && finfo.completeSuffix().contains( pattern.mid( pos + 1 ), cs ) )
        return true;

    // test folders .../....
    pos = pattern.indexOf( sep );
    if  ( pos >= 0 && finfo.path().contains( pattern, cs ) )
        return true;

    // test raw name pattern
    if  ( finfo.fileName().contains( pattern, cs ) )
        return true;

    return false;
}

void RecollAsyncWorker::postProcess() {
    QStringList nullList;
    recollparser* parser = new recollparser( nullList );  //null version
    
    if  ( m_options->m_bRunRecoll ) {
        Plasma::QueryMatch myMatch( m_runner );
        myMatch.setType( Plasma::QueryMatch::HelperMatch );
        
        QString wholeCmd = RECOLLCMD;
        wholeCmd += " ";
        wholeCmd += m_recollArgs.join( " ");
        wholeCmd.remove( RECOLLCLI );
        myMatch.setText( "Run recoll: " + m_term );
        myMatch.setSubtext( wholeCmd );
        
        QString cmd = EXECCMD;
        cmd += RECOLLCMD;
        cmd += ' ';
        //cmd += " \"";
        cmd += m_recollArgs.join( " " ).remove( RECOLLCLI );
        //cmd += '"';

        QStringList cmdData;
        cmdData << cmd;
        cmdData << parser->data();
        myMatch.setId( QString( -1 ) );
        myMatch.setData( cmdData );
        myMatch.setIcon( m_icon );
        myMatch.setEnabled( true );
        myMatch.setRelevance( RELEV_EXEC );
        
        m_context.addMatch( m_term, myMatch );
    }
    if  ( m_options->m_bRunKIO ) {
        Plasma::QueryMatch myMatch( m_runner );
        myMatch.setType( Plasma::QueryMatch::HelperMatch );
        
        QString wholeCmd = DOLPHINCMD;
        wholeCmd += " ";
        wholeCmd += KIO_RECOLLCMD;
        wholeCmd += '/';
        wholeCmd += m_term;
        wholeCmd += '/';
        myMatch.setText( "Run " + wholeCmd );
        myMatch.setSubtext( wholeCmd );
        
        QString cmd = EXECCMD;
        cmd += DOLPHINCMD;
        cmd += " \"";
        cmd += KIO_RECOLLCMD;
        cmd += '/';
        cmd += m_term;
        cmd += "/\"";

        QStringList cmdData;
        cmdData << cmd;
        cmdData << parser->data();
        myMatch.setId( QString( -2 ) );
        myMatch.setData( cmdData );
        myMatch.setIcon( m_icon ); // KIcon( "dolphin" ) );
        myMatch.setEnabled( true );
        myMatch.setRelevance( RELEV_EXEC );
        
        m_context.addMatch( m_term, myMatch );        
    }
    
}

// for debugging
void RecollAsyncWorker::echoCmd( QString cmd, QStringList arg ) {    
        Plasma::QueryMatch myMatch( m_runner );
        myMatch.setType( Plasma::QueryMatch::NoMatch );
        myMatch.setSubtext( "DEBUG" );
        QString wholecmd = cmd;
        wholecmd += " ";
        wholecmd += arg.join( " ");
        myMatch.setText(wholecmd);   
        myMatch.setIcon( m_icon );
        m_context.addMatch( m_term, myMatch );
}

void RecollAsyncWorker::echoStr( QString str ) {
        Plasma::QueryMatch myMatch( m_runner );
        myMatch.setType( Plasma::QueryMatch::NoMatch );
        myMatch.setSubtext( "DEBUG" );
        myMatch.setText(str);
   
        myMatch.setIcon( m_icon );
        m_context.addMatch( m_term, myMatch );
    
}


#include "recollworker.moc"